/*
 * Higher level(?) Modem control routines.
 *
 * Copyright 1999          Pavel Machek
 * Copyright 1999          Jamie Lokier
 * Copyright 1999-2000     R.J.M. Close
 *
 * Can be distributed under GPL
 */

#include "control.h"

void go_offhook(void)	/* Ouch, is it real command for going offhook? It looks like it is used only in some tests! */ 
{
  printf( "Going offhook..." );
  x_output(0x13);
  x_output(0x11);
  x_output(0x02);
  dp_state = 0x14;
  printf( "ok\n" );
}

// Command to actually go offhook.
// Based on Jamie's version with minor changes.
bool dp_offhook_cmd (void)
{
  if (cell_active)
    {
      at_init_diagnostics ();
      x_status[0] = 0;
      return true;
    }

  // Wait until ring not detected or 3 seconds has expired.
  // Basically wait up to 3 seconds for line to go quiet.
  dp_timer = x_current_time();
  do {
    if (dp_ring_detected() == false) break;
	printf ("~");
  }
  while (x_elapsed_time(dp_timer) < 3000);

  x_output(14);
  x_output(17);
  x_output(2);

  if (H_shunt_relay_on_time != 0xff)
    x_output (15);
 
  if (S_speaker_enable) {
    switch (S_speaker_volume) {
    case 3: x_output(0x0b); break;
    case 2: x_output(0x0a); break;
    default: x_output(0x09); break;
    }
    dp_modem_command(0x13, 0, 0);
    x_output(8);
  }

#if !PAVEL_FIX
  if (H_overcurrent_sense_validation_time != 0
     && dp_dialing_status == 0)
    {
      unsigned local_timer;
      x_output (28);
      x_output (30);
	  
      START_TIMER (dp_timer);
	  local_timer = dp_timer;
	  while (ELAPSED_TIME (local_timer) <= 10)
		{
		  if (x_input (65) == 0)
			continue;
		  START_TIMER (local_timer);
		  if (ELAPSED_TIME (dp_timer)
			  >= (H_overcurrent_sense_validation_time * 10))
			{
			  x_output (3);
			  x_send_mail (7, 1, 0, 0);
			  dp_over_current_alert ();
			  return false;
			}
		}

      if (H_DC_loop_limiting_relay_flag != 0)
		x_output (27);
      if (H_DC_loop_VI_characteristics_relay_flag != 0)
		x_output (29);
    }
#endif

  if (x_modem_mode != 2 && x_modem_mode != 3)
    {
      x_output (19);
      at_init_diagnostics ();
      x_status[0] = 0;
    }
  else
    {
      x_output (20);
      if (CpqFlag) x_output (13);
    }

#if !PAVEL_FIX
  if (H_shunt_relay_on_time != 0xff)
    {
      int limit = H_shunt_relay_on_time * 10;
      START_TIMER (dp_timer);
      if (H_overcurrent_sense_validation_time != 0 && limit >= 10)
		limit -= 10;
      while ((dp_dialing_status == 0)
             && (ELAPSED_TIME (dp_timer) <= limit))
		/*empty*/;
      x_output (16);
    }
#endif

  dp_init_local_phone_timer ();

  return true;
}

void dp_hdlc_mode(void)
{
  DpIdleFillChar = 0x7e;
  dp_modem_command_long(0x29, 0x7e, 1, 0x40, 0);
}

void dp_onhook_cmd(void)
{
  printf( "Going onhook: " );
  //  printf( "Should call cell_onhook()\n" );
  dp_modem_command(0x13, 0, 0);
  x_output(   7);
  x_output(0x0e);
  x_output(0x14);
  x_output(0x10);
  if (S(0x72) == 0) { 
    x_output(0x19);
    x_output(   3);
    x_output(0x12);
  } else
	S(0x72) = 0x0a;

  dp_write_dsp_ram(0xf20, 0x42f8);
  dp_regwrite(0xd8, 0xff);
  dp_regwrite(0xb7, 0xff);

  dp_tad_downloaded = dp_fax_downloaded = 0; 
  if (S(0x74) == 1) {
    S(0x74) = 0;
	S(0x52) = 0;
  }
  v8bis_app_reset();
  dp_init_local_phone_state();
  V34Mode = 0;
  dp_update_diagnostics();
  if (x_status[0] != 0)
    goto l2D1;
  x_status[0] = 1;
  //  ll_save_diagnostics();

 l2D1:
  if ((S(0x72) == 0) && (x_modem_mode != 3))
    dp_run_rom();

  dp_first_call = 0;
  printf( "\n" );
}

bool dp_carrier_loss(void)
{
  unsigned char retVal;

  // Read value from DSP RAM.
  retVal = dp_read_dsp_ram(0x8E);
  // Invert it.
  retVal = ~retVal;

  // Return bit 1 state as bool.
  retVal = ((retVal >> 1) & 1);

  if (retVal) {
	printf ("Carrier lost.\n");
    return true; 
  }
  else {
	printf ("Carrier present.\n");
	return  false;
  }
}

unsigned char dp_retrain(void)
{
  return ((dp_dsp_status >> 4) & 1);
}

// Sets the new modem rate.
void dp_modem_rate(void)
{
unsigned short new_line_rate;

  dp_dsp_status = (0xff & dp_regread(0x32));
  if (dp_dsp_status & 0x10) return;

  new_line_rate = (dp_dsp_status & 0xf) + 2;

  // If new line is same as old then just bugger off.
  if (new_line_rate == x_line_rate) return;

  if (new_line_rate < x_line_rate)
	dp_bFallForward = 1;
  else {
	if (dp_bFallForward != 2)
	  dp_bFallForward = 0;
	else
	  dp_bFallForward = 1;
  }
  //32d90:	
  if (V34Mode) return;
#if LT_DEBUG
  printf("dp_modem_rate: new rate set to: %d.\n", new_line_rate);
#endif
  x_line_rate = new_line_rate;
  x_line_rate_tx = new_line_rate;
  dp_bGoodAuto = 0;
}

void dp_async_8bit_mode(void)
{
  DpIdleFillChar = 0xff;
  dp_modem_command_long(0x29, 0xff, 1, 0x87, 0);
}

// Set speaker up acording to values held in globals:
// speakerVolume and speakerEnable.
void dp_56k_speaker_cmd (void)
{ 
  if (!S_speaker_enable)
	{
	  dp_write_dsp_ram(0xD, 0);
	  return;
	}

  if (S_speaker_volume == 3) 
	{
	  x_output(0xB);
	  x_output(8);
	  return;
	}

  if (S_speaker_volume == 2)
	{
	  x_output(0xA);
	  x_output(8);
	  return;
	}
  // speakerVolume = 1, or not valid.
  x_output(9);
  x_output(8);
  return;

}


// Fill in, if needed.
unsigned short dp_test_cmd (unsigned int arg1, unsigned char* arg2)
{
  // Crude dummy to use args. Code does not do this!
  *arg2 = arg1;

  printf ("dp_test_cmd: not yet implemented!\n");

  return 1;
}


void Modem_Resume(void)
{ 
   x_send_mail (0x66, 1, 0, 0);
   x_wakeup();
}

void dp_download_dsp (unsigned char* data, unsigned count)
{
  unsigned char * local_ptr = (unsigned char *) data;
  unsigned char save_rd7 = dp_bamil_rd7;

  if (x_dsp_mars == 1)
    {
      Write_mdm_word (0x3c, 0x1ff);
      Write_mdm_word (0x3e, 0);
    }
  dp_bamil_rd7 = 0xff;
  dp_regwrite (0xd8, 0xff);
  dp_regwrite (0xd7, 0xff);
  dp_modem_command_long (0x29, 0, 0, 0, 0);
  dp_write_dsp_ram (0x39, 0);

  while (count != 0)
    {
      unsigned block_size = *(unsigned short *) (local_ptr + 2);
      unsigned counter;

	  printf("count %x, ", count);

      count -= (3 + block_size);
      counter = 0;
      dp_regwrite (0x30, 1);

      do
		dp_regwrite (0x31 + counter, *local_ptr++);
      while (++counter < 6);

      dp_regwrite (0xd8, 0x18);
      dp_regwrite (0x37, 0x1a);
      wait_for_core_read ();

      while (block_size > 0)
		{
		  unsigned tmp;

		  tmp = block_size >= 3 ? 3 : block_size;
		  block_size -= tmp;
		  tmp += tmp;
		  counter = 0;
		  do
			dp_regwrite (0x30 + counter, *local_ptr++);
		  while (++counter < tmp);
		  dp_regwrite (0xd8, 0x18);
		  dp_regwrite (0x37, tmp | (block_size ? 0xc0 : 0xc8));
		  wait_for_core_read ();
		}
    }

  dp_regwrite (0xb7, 0xff);
  dp_regwrite (0xd8, 0xff);
  dp_bamil_rd7 = save_rd7;
  dp_regwrite (0xd7, dp_bamil_rd7);
  if (x_dsp_mars == 1)
    {
      Write_mdm_word (0x3c, 0x1ff);
      Write_mdm_word (0x3e, 1);
    }
}

void dp_download_fax(void)
{
  //printf( "Download fax: size is %x\n", sizeof(wDspFax));
  dp_modem_command(0x13, 0, 0);
  //dp_download_dsp(wDspFax, 0x7FA);
  dp_modem_command(0x13, 0, 0);
 }

void dp_retrain_cmd(void)
{
  dp_modem_command(0x0F, 0, 0);
}

void dp_fast_retrain_cmd(void)
{
  dp_lapm_error_check_init();
  dp_update_rate();
  dp_modem_command(2, 0, 0);
}

void dp_originate_cmd (void)
{
  printf("dp_originate_cmd called.\n");
  dp_modem_command(0x13, 0, 0);
  dp_update_modem_options();
  dp_modem_command(0x0e, 1, 0);
  dp_v8_originate_msg();
}

void dp_answer_cmd(void)
{
  dp_modem_command(0x13, 0, 0);
  dp_update_modem_options();
  dp_modem_command(0x0e, 0, 0);
}

bool dp_line_connect(void)
{
  if (dp_regread(0x19) != 0x10) return true;

  if (dp_regread(0x16) != 4) return true;

  return false;
}


// Update the data rate.
void dp_update_rate(void)
{

  unsigned char bl = 0;
  unsigned int ebp = 0;
  unsigned int esi = 0;
  unsigned int eax = 0;
  unsigned int edi = 0;
  unsigned short ax;
  unsigned char al;

  //printf("dp_update_rate: setting up data rate...");

  wSymbolRate = 0x3FF;
  //edx = 5;
  if (S(0x6f) == 0) goto loc_2C870;

  al = S(0x10) - 1;
  if (al <= 8) {
    switch (al){
    case 0: case 1: case 2:
	bl = 3;
	break;
    case 3:
	bl = 5;
	break;
    case 4:
	bl = 6;
	break;
    case 5:
	bl = 7;
	break;
    case 6:
	bl = 9;
	break;
    case 7:
	bl = 11;
	break;
    case 8:
	bl = 13;
	break;
    }
  }
loc_2C870:
  al = S_dial_line_rate;
  if (bl == 0) goto loc_2C881;
  if (al == 0) goto loc_2C883;
  if (bl <= al) goto loc_2C883;

loc_2C881:
  bl = al;

loc_2C883:
  al = bl;
  al = al - 2;
  //ecx = 0x10;
  if (al > 16) goto loc_2CB1A; // Out of range. (same as case 2)

  switch (al) {
  case 0:
    // loc_2C89B 0
    edi = 0x80;
    //goto loc_2C8DF;
    // loc_2C8DF:
    x_line_rate = 4;
    break;

  case 1:
    // loc_2C8A2
    if ((S(0x3a) == 0) && (country_code == COUNTRY_FRANCE)) {
      edi = 0x20;
      eax = 3 + H_data_or_fax_carrier_rx_threshold_dB;
      dp_write_dsp_ram(0x37, eax);
      } 
    else
      edi = 0x10;

    //loc_2C8CD:	
    x_line_rate = 2;
    break; //goto loc_2CB9B;

  case 2: 
	// loc_2CB1A 2
  loc_2CB1A:
    al = S(0x38);
	if ((al == 8) || (al == 10) || (al == 0) || (al == 1)) {
	  //loc_2CB72: //Same as case: 16
	  ebp = 0xFFF0;
	  if (S_auto_rate)
		eax = 0xffffffff;    // -1
	  else
		eax = 0;

	  edi = 0xb1;
	  eax = eax & 0x8FFF;
	  x_line_rate = 17;
	  eax = eax + 0x1000;
	  esi = eax;
	  break; //goto loc_2CB9B;
	}

	wSymbolRate = 0x07FF;
	if (S_auto_rate == 0){
	  ebp = 0xC990;
	  esi = 0x2000;
	}
	else {
	  //loc_2CB4D:
	  ebp = 0xFFF0;
	  esi = 0xBFFF;
	}
	//loc_2CB57: 
	edi = 0xb1;

	// S_V34_enable = S(0x1c) in 568 dis? (1c = 28)
	// 568 code extract...
	if (S_V34_enable)
	  x_line_rate = 18;
	else
	  x_line_rate = 12;

	// Replaces...
	//  S_V34_enable = al;
	// neg	 al
	// sbb	 eax, eax  ( if al eax = fffffff else 0)
	// and	 eax, 6
	// eax, 0Ch
	// mov	 dword__x_line_rate, eax
	break;


  case 3:
	// loc_2C8DC 3
	edi = 2;

  //loc_2C8DF:
	x_line_rate = 4;
	break; //goto  loc_2CB9B;

  case 4: 
	//loc_2C8EE 4
	//al = S_auto_rate;    //byte_59E8A

	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 1;
	eax = eax & 0x8000;
	x_line_rate = 5;      // 5.
	eax++;
	wSymbolRate = edi;      // 1.
	esi = eax;
	break; //goto loc_2CB9B;

  case 5:
	// loc_2C914 5
	//al = S_auto_rate;
	ebp = 0x89B0;

	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;
	//0002C91E		 neg	 al
	//0002C920		 sbb	 eax, eax

	edi = 0x0B1;
	eax = eax & 0x8001;
	x_line_rate = 6;
	eax = eax + 2;
	esi = eax;
	break; //goto loc_2CB9B;

  case 6:
	//loc_2C93F   (32011: in Jamie's dis.) 
      if (S_auto_rate == 0){
        ebp = 0x8B90;
        esi = 4;
        }
      else {
        //loc_2C952:
        ebp = 0x8BB0;
        esi = 0x8007;
        }
   //loc_2C95C:
      edi = 0xb1;
      x_line_rate = 7;
      break; //goto loc_2CB9B;

  case 7:
	//loc_2C970 7
      if (S_auto_rate == 0){
        ebp = 0x89D0;
        esi = 8;
        }
      else {
        //loc_2C983:
        ebp = 0x8BF0;
        esi = 0x800F;
        }

      // loc_2C98D:
      edi = 0xb1;
      x_line_rate = 8;
      break; //goto loc_2CB9B;

  case 8:
	// loc_2C9A1 8
      if (S_auto_rate == 0){
        ebp = 0x89D0;
        esi = 0x10;
        }
      else {
        //loc_2C9B3:
        ebp = 0x8FF0;
        esi = 0x801F;
        }

      // loc_2C98D:
      edi = 0xb1;
      x_line_rate = 9;
      break; //goto loc_2CB9B;

  case 9:
	// loc_2C9D1 9
      if (S_auto_rate == 0){
        ebp = 0x9990;
        esi = 0x20;
        }
      else {
        //loc_2C9E4:
        ebp = 0x9FF0;
        esi = 0x803F;
        }
      // loc_2C9EE
      edi = 0xb1;
      x_line_rate = 10;
      break; //goto loc_2CB9B;

  case 10:
	// loc_2CA02 10
      if (S_auto_rate == 0){
        ebp = 0x0A990;
        esi = 0x40;
        }
      else {
        //loc_2C9E4:
        ebp = 0xBFF0;
        esi = 0x807F;
        }
      // loc_2C9EE
      edi = 0xb1;
      x_line_rate = 11;
      break; //goto loc_2CB9B;

  case 11:
	//loc_2CA33 11
      if (S_auto_rate == 0){
        ebp = 0x0C990;
        esi = 0x80;
        }
      else {
        //loc_2CA48:
        ebp = 0xFFF0;
        esi = 0x80FF;
        }
      // loc_2CA52
      edi = 0xb1;
      x_line_rate = 12;
      break; //goto loc_2CB9B;

  case 12:
	//loc_2CA66 12
	ebp = 0xFFF0;
	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 0xb1;
	eax = eax & 0x80FF;
	x_line_rate = 13;
	eax = eax + 0x100;
	esi = eax;
	break; //goto loc_2CB9B;

  case 13:
	//loc_2CA94 13
	ebp = 0xFFF0;
	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 0xb1;
	eax = eax & 0x81FF;
	x_line_rate = 14;
	eax = eax + 0x200;
	esi = eax;
	break; //goto loc_2CB9B;

  case 14:
	// loc_2CAC2 14
	ebp = 0xFFF0;
	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 0xb1;
	eax = eax & 0x83FF;
	x_line_rate = 15;
	eax = eax + 0x400;
	esi = eax;
	break; //goto loc_2CB9B;

  case 15:
	// loc_2CAF0 15
	ebp = 0xFFF0;
	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 0xb1;
	eax = eax & 0x87FF;
	x_line_rate = 16;
	eax = eax + 0x800;
	esi = eax;
	break; //goto loc_2CB9B;

  case 16:
	// loc_2CB72 16
	ebp = 0xFFF0;
	if (S_auto_rate)
	  eax = 0xffffffff;    // -1
	else
	  eax = 0;

	edi = 0xb1;
	eax = eax & 0x8FFF;
	x_line_rate = 17;
	eax = eax + 0x1000;
	esi = eax;
	break; //goto loc_2CB9B;
  }	
  //end switch.

// This stuff is very different to what's in Jamie's and
// the 568 version!
//loc_2CB9B:
  if (S_auto_mode) goto loc_2CBAF;
  if (bl == 0) goto loc_2CBAD;
  if (bl < 7) goto loc_2CBAF;

loc_2CBAD:
  edi = 0;

loc_2CBAF:
  //ebx = 0x821;
  if (ETC_mode == 0) goto loc_2CC2E;
  if (ETC_mode & 0x80) goto loc_2CBF6;

  al = S(0x28);
  if (al != 2) goto loc_2CBDD;

  ebp = ebp & 0x8BF0;
  esi = esi & 0x800F;
  goto loc_2CBED;

loc_2CBDD:	
  if (al != 1) goto loc_2CBED;
  ebp = ebp & 0x89B0;
  esi = esi & 0x8003;

loc_2CBED:
  ETC_mode = ETC_mode & 0x80;
  goto loc_2CC0C;

loc_2CBF6:
  ebp = ebp | 0x9ff0;
  esi = esi | 0x83ff;

  S_0x5d = dp_read_dsp_ram(0x821);

loc_2CC0C:
  if (wSymbolRate != 1)
    wSymbolRate = S(0x64);	

//loc_2CC22:	
  dp_write_dsp_ram(0x824, 0);

loc_2CC2E:
  if ((S(0x75) & 0xf0) == 0x30)
     edi = edi & 0xFF7F;
 
//loc_2CC3F: // Similar to 32292: in Jamies dis.	
  dp_write_dsp_ram(1, ebp);
  dp_write_dsp_ram(2, edi);

  ax = S(0x5D);
  if ((0xffff & edi) != 0) 
    ax = ax | 8;
  else //loc_2CC5D:	
    ax = ax & 0xfff7;

//loc_2CC61:
  dp_wV34Option = ax & 0xefff;

  if (S_V34_enable == 0) {
    //loc_2CCEB:
    dp_write_dsp_ram(0x80D, 0);
    goto loc_2CCED;
    }
  if (x_modem_state != 7) goto loc_2CCB8;
  if ((esi & 0x3000) == 0) goto loc_2CC9F;

  wSymbolRate = 0x400;
  ax = ax | 0x800;
  esi = esi | 0x800;
  dp_wV34Option = ax;
  goto loc_2CCB8;

loc_2CC9F:
  if (esi == 1) 
    dp_wV34Option = 0;	
  else //loc_2CCAF:
    wSymbolRate = 0x100;

loc_2CCB8:	
  if (wSymbolRate & 0x400)
    dp_write_dsp_ram(0x810, 0x21FF);
  else //loc_2CCC8:	
    dp_write_dsp_ram(0x810, 0x217F);

  dp_write_dsp_ram(0x823, wSymbolRate);

  dp_write_dsp_ram(0x80D, esi);

loc_2CCED:
  dp_write_dsp_ram (0x821, dp_wV34Option);
  dp_write_dsp_ram (0x3c, 0x122);
  dp_write_dsp_ram (0x47, 0x123);

  //printf(" xSymbolRate: 0x%04x, x_line_rate: %d.\n",
  //       wSymbolRate, x_line_rate);

}

void dp_update_modem_options(void)
{
  unsigned int bl, ax, esi, edi = 0;

  if (x_modem_mode != 0) return;

  //printf("dp_update_modem_options: setting options in DSP RAM.\n");

  bl = 0x27;
loc_2CD9D:
  ax = bl;
  ax = ax + 0x800;
  dp_write_dsp_ram(ax, edi);
  ax = bl;
  bl--;
  if (ax != 0) goto loc_2CD9D;

  dp_write_dsp_ram(0x0c, edi);
  dp_write_dsp_ram(0x39, edi);
  dp_write_dsp_ram(0x48, 0x7d0);

  ax = H_data_or_fax_carrier_rx_threshold_dB; // 59f26
  dp_write_dsp_ram (0x37, ax);

  esi = 0;

  if (cell_active) {
    dp_write_dsp_ram(0x37, 0x2b);
    esi = 0x4000;
  }

  ETC_mode = ETC_mode & 0x7f;
  dp_update_rate();

  ax = S_dial_line_rate;
  ax--;
  ax--;
  if (ax == 0) goto loc_2CE24;
  ax = ax - 3;
  if (ax != 0) goto loc_2CE30;

  if (S(0x3a) == 1) esi = esi | 1;
  goto loc_2CE30;

loc_2CE24:
  if (S(0x3b) != 0) esi = esi | 0x10;

loc_2CE30:
  ax = S(0x35);
  ax--;
  if (ax == 0) goto loc_2CE42;
  ax--;
  if (ax != 0) goto loc_2CE45;

  esi = esi | 4;
  goto loc_2CE45;

loc_2CE42:
  esi = esi | 2;

loc_2CE45:
  dp_write_dsp_ram(0x38, esi);
  dp_async_user_mode();

  esi = 0x825;
  dp_write_dsp_ram(esi, 0x3000);
  bl = S(0x54);
  ax = S(0x38);
  bl = bl/2;
  bl = bl & 0x70;

  if ((ax < 2) || (ax > 7))
	dp_write_dsp_ram(esi, 0x2333);
  else
	if (bl == 0x60)
	  bl = bl + 4;

  //loc_2CE9F:
  ax = bl;
  dp_write_dsp_ram(0x822, ax);
  dp_write_dsp_ram(0x827, 2);

  if ((S_dial_line_rate != 2) | (S(0x3b) != 1) | (x_modem_state != 3))
	ax = H_transmit_level_ndB;
  else
	ax = H_transmit_level_ndB + 1;

  dp_transmit_level(ax);
  dp_write_dsp_ram(0x3a, edi);
  dp_bV34WkState = 1;

}

void dp_update_diagnostics (void)
{
  if (V34Mode != 0)
    dp_get_line_rate ();
  x_status[6]  = x_line_rate_tx;
  x_status[8]  = x_line_rate;
  x_status[0x10] = dp_read_dsp_ram (0x4c);
  x_status[0x0c]  = x_modem_ec;
  x_status[0x12] = dp_read_dsp_ram (0x4a);
  x_status[0x14] = dp_read_dsp_ram (0x03);
  x_status[0x16] = dp_read_dsp_ram (0x54);
  x_status[0x18] = dp_read_dsp_ram (0x55);
  x_status[0x1a] = dp_read_dsp_ram (0x56);
  x_status[0x32] = dp_read_dsp_ram (0x80a);
  x_status[0x34] = dp_read_dsp_ram (0x80b);
  x_status[0x38] = x_dial_residual;
}

void dp_get_line_rate (void)
{
  unsigned short word = dp_read_dsp_ram (0x36);
  x_line_rate_tx = ((word >> 8) & 0x1f) + 2;
  x_line_rate    = (word & 0x1f) + 2;
}

void dp_v8_originate_msg(void)
{
  printf("dp_v8_originate_msg: warning not yet implemented!\n");
}

// Taken from version 568.
void dp_disable_workaround(void)
{
  if (x_dsp_mars3 != 0) {
	// Mars3 chipset.
	dp_write_dsp_ram(0xf20, 0x4228);
	dp_write_dsp_ram(0x311, 0xb9a7);
  }
  else {
	dp_write_dsp_ram(0xf20, 0x42f8);
	dp_write_dsp_ram(0x311, 0xb994);
  }
}

void dp_transmit_level (unsigned int level)
{
  dp_write_dsp_ram(3, level);
}

void dp_async_user_mode(void)
{
  unsigned int var_4 = 0x080;

  switch(S(13)){
  case 5:
	break;
  case 6:
	var_4 = 0x82;
	break;
  case 7:
	var_4 = 0x84;
	break;
  default:
	var_4 = 0x86;
	break;
  }

  //loc_2C7D2:
  if (S(15) != 0)
	var_4 = var_4 | 1;

  switch (S(0x13)) {
  case 1:
	var_4 = var_4 | 0x38;
	break;
  case 2:
	var_4 = var_4 | 0x30;
	break;
  case 3:
	var_4 = var_4 | 1;
	break;
  case 4:
	var_4 = var_4 | 0x28;
	break;
  default:
	break;
  }
  // loc_2C808:
  DpIdleFillChar = 0xff;
  dp_modem_command_long(0x29, 0xff, 1, var_4, 0);
}

void dp_sleep_mode(unsigned char arg1)
{
  printf("dp_sleep_mode(%d): warning not yet implemented!\n", arg1);
}

void dp_over_current_alert (void)
{
  printf("dp_over_current_alert: warning not yet implemented!\n");
}

// I wonder what this does? ;)
void dp_disable_33600 (void)
{
  dp_write_dsp_ram(0x801, 0x217F);
  dp_write_dsp_ram(0x823, wSymbolRate & 0x0FBFF);
  dp_write_dsp_ram(0x80D, (dp_read_dsp_ram(0x80D) & 0x0DFFF));
}

void dp_v34_workaround(void)
{
  printf("dp_v34_workaround: warning not yet implemented!\n");
}

void
dp_enable_ring_int(void)
{
  dp_regwrite(0xd8, 0xff);
  dp_bamil_rd7 &= 0xf1;
  dp_regwrite(0xd7, dp_bamil_rd7);
}

void dp_modem_idle_cmd (void)
{
  dp_modem_command (0x13, 0, 0);
  dp_write_dsp_ram (0xf20, 0x42f8);
}

bool dp_init_modem_connection(void)
{
  unsigned int ecx;
  unsigned short ax;

  dp_count = 0;
  if (V34Mode) {
	dp_get_line_rate();   // Set line rate var's from value read from DSP.
	if (x_line_rate == 2) return false;
	
	ax = (V34Mode & dp_read_dsp_ram(80));
	x_status[1] = 2;
	dp_max_line_rate = 18;

	//loc_2D43C: // Loop
	while ((ax & 0x2000) == 0){
	  ecx = dp_max_line_rate;
	  dp_max_line_rate--;
	  if (ecx == 0) break;
	  ax = ax << 1;
	}
	//goto loc_2D49E;
  }
  else{
	//loc_2D455:
	dp_max_line_rate = x_line_rate;
	dp_modem_rate();
	x_status[1] = 1;
	x_line_rate_tx = x_line_rate;
	
	if ((S_auto_mode == false) && (S_auto_rate == 0)
		&& (dp_max_line_rate != x_line_rate)) return 0;

	//loc_2D491:	
	dp_wV32bisRcvRate = dp_read_dsp_ram(0x35);
	//goto loc_2D49E;
  }

  //loc_2D49E:
  dp_bFallForward = 1;
  if (x_line_rate >= 6) 
	dp_write_dsp_ram(0xA60, 0);

  //loc_2D4B9:
  if ((S_speaker_enable == 1) || 
	  (S_speaker_enable == 3))
	x_output(7);

  //loc_2D4CD:
  if (x_modem_mode != 0) goto loc_2D505;
  if (S_dial_line_rate == 2) goto loc_2D4FE;
  if (dp_read_dsp_ram(0x3b) == 9) goto loc_2D4FE;

  S(0x1b) = 0;
  if (dp_read_dsp_ram(0x3b) != 0x86) goto loc_2D505;

loc_2D4FE:
  S(0x1b) = 1;

loc_2D505:
  if (((S(0x53) == 5) || (S(0x55) == 6))
	  && (!V34Mode)) return false;

  //loc_2D524:
  if ((wSymbolRate & 0x400) != 0)  // Test bit 4 of MSB.
	dp_disable_33600();

  //loc_2D532:
  x_status[0x2] = (x_line_rate_tx & 0xff);
  x_status[0x3] = (unsigned char)((x_line_rate_tx & 0xff00)/256);
  x_status[0x4] = (x_line_rate & 0xff);
  x_status[0x5] = (unsigned char)((x_line_rate & 0xff00)/256);
  dp_update_diagnostics();

  return true;
}

// ** WARNING! the Lucent 568 version is very different and has some
// DSP code patches (RT_FIX_MARS3 or RT_FIX) depending on whether it's
// a mars3 chipset. See ll_set_pci_id in 568 version for device specific
// settings.
bool dp_dial_connect_success(void)
{

  printf("dp_dial_connect_success: called\n");
  dp_v34_workaround();

  if ((dp_read_dsp_ram(0x8e) & 2) == 0) goto loc_2D620;
  if ((dp_read_dsp_ram(0x3a) & 8) == 0) goto loc_2D620;

  dp_dsp_status = (unsigned char)(dp_read_dsp_ram(0x32) & 0xff);
  if ((dp_dsp_status & 0x10) != 0) goto loc_2D620;

  dp_bGoodAuto = false;
  dp_bFallForward = true;

  V34Mode = dp_read_dsp_ram(0x817);
  if (V34Mode){
	if (dp_read_dsp_ram(0x80) != 0x70EA) goto loc_2D620;
	if (dp_read_dsp_ram(0xCF) != 0x6B69) goto loc_2D620;
  }

  if (ETC_mode){
	dp_count = 0;

	if (V34Mode)
	  dp_write_dsp_ram(0x80D, (dp_read_dsp_ram(0x80D) & 0x8003));
	else
	  dp_write_dsp_ram(1, (dp_read_dsp_ram(1) & 0x89B0));
  }

  dp_lapm_error_check_init();
  return dp_init_modem_connection();

 loc_2D620:
  // Debug 
  printf("dp_dial_connect_success: failed!\n");
  return false;
}

void dp_rate_change_cmd(unsigned int rate)
{
  // 568 version, rate arg is not used!
  sprintf(NULL, "%d", rate); // Bodge to consume arg.

  dp_disable_workaround();
  dp_bGoodAuto = 0;
  dp_count = 0;
  *(unsigned short*)(x_status + 0x24) = *(unsigned short*)(x_status + 0x24) + 1;
  dp_lapm_error_check_init();

  if ((x_dsp_mars3 != 4) &&
	  (dp_read_dsp_ram(0x60) > 16))
	dp_transmit_level(18);

  dp_modem_command(15, 0, 0);
}

void dp_auto_rate (void)
{
  printf("dp_auto_rate: Warning not yet implemented!\n");
}
